import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    route_table_id: str,
    resource_id: str = None,
    destination_cidr_block: str = None,
    destination_ipv6_cidr_block: str = None,
    destination_prefix_list_id: str = None,
    vpc_endpoint_id: str = None,
    egress_only_internet_gateway_id: str = None,
    gateway_id: str = None,
    instance_id: str = None,
    nat_gateway_id: str = None,
    transit_gateway_id: str = None,
    local_gateway_id: str = None,
    carrier_gateway_id: str = None,
    network_interface_id: str = None,
    vpc_peering_connection_id: str = None,
    core_network_arn: str = None,
) -> Dict[str, Any]:
    """Creates a route in a route table within a VPC. You must specify one of the following targets: internet gateway
    or virtual private gateway, NAT instance, NAT gateway, VPC peering connection, network interface, egress-only
    internet gateway, or transit gateway. When determining how to route traffic, we use the route with the most
    specific match. For example, traffic is destined for the IPv4 address 192.0.2.3, and the route table includes
    the following two IPv4 routes:    192.0.2.0/24 (goes to some target A)    192.0.2.0/28 (goes to some target B)
    Both routes apply to the traffic destined for 192.0.2.3. However, the second route in the list covers a smaller
    number of IP addresses and is therefore more specific, so we use that route to determine where to target the
    traffic. For more information about route tables, see Route tables in the Amazon Virtual Private Cloud User
    Guide.

    Args:
        name(str): An Idem name of the resource.
        route_table_id(str): The ID of the route table for the route.
        resource_id(str, Optional): Unique identifier to identify the route in a route table. format of resource id is
            <route_table_id>/<destination_cidr_block or destination_ipv6_cidr_block or destination_prefix_list_id>
        destination_cidr_block(str, Optional): The IPv4 CIDR address block used for the destination match. Routing decisions are based on the
            most specific match. We modify the specified CIDR block to its canonical form; for example, if
            you specify 100.68.0.18/18, we modify it to 100.68.0.0/18. Defaults to None.
        destination_ipv6_cidr_block(str, Optional): The IPv6 CIDR block used for the destination match. Routing decisions are based on the most
            specific match. Defaults to None.
        destination_prefix_list_id(str, Optional): The ID of a prefix list used for the destination match. Defaults to None.
        vpc_endpoint_id(str, Optional): The ID of a VPC endpoint. Supported for Gateway Load Balancer endpoints only. Defaults to None.
        egress_only_internet_gateway_id(str, Optional): [IPv6 traffic only] The ID of an egress-only internet gateway. Defaults to None.
        gateway_id(str, Optional): The ID of an internet gateway or virtual private gateway attached to your VPC. Defaults to None.
        instance_id(str, Optional): The ID of a NAT instance in your VPC. The operation fails if you specify an instance ID unless
            exactly one network interface is attached. Defaults to None.
        nat_gateway_id(str, Optional): [IPv4 traffic only] The ID of a NAT gateway. Defaults to None.
        transit_gateway_id(str, Optional): The ID of a transit gateway. Defaults to None.
        local_gateway_id(str, Optional): The ID of the local gateway. Defaults to None.
        carrier_gateway_id(str, Optional): The ID of the carrier gateway. You can only use this option when the VPC contains a subnet which
            is associated with a Wavelength Zone. Defaults to None.
        network_interface_id(str, Optional): The ID of a network interface. Defaults to None.
        vpc_peering_connection_id(str, Optional): The ID of a VPC peering connection. Defaults to None.
        core_network_arn(str, Optional): The Amazon Resource Name (ARN) of the core network. Defaults to None.

    Request Syntax:
       .. code-block:: sls

          [route-table-id-destination]:
            aws.ec2.route.present:
              - name: "string"
              - route_table_id: "string"
              - destination_cidr_block: "string"
              - gateway_id: "string"
              - resource_id: "string"

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            r-rtb-343452242424:
              aws.ec2.route.present:
                - name: r-rtb-343452242424
                - route_table_id: rtb-3434522
                - resource_id: r-rtb-343452242424
                - destination_cidr_block: 172.12.12.201
                - gateway_id: subnet-4923930130
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    if resource_id:
        before = await hub.exec.aws.ec2.route.get(
            ctx=ctx, name=name, resource_id=resource_id
        )
        if not before["result"] or not before["ret"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result
        result["old_state"] = copy.deepcopy(before["ret"])
        plan_state = copy.deepcopy(result["old_state"])
        update_allowed_params = {
            "carrier_gateway_id": "CarrierGatewayId",
            "egress_only_internet_gateway_id": "EgressOnlyInternetGatewayId",
            "gateway_id": "GatewayId",
            "instance_id": "InstanceId",
            "local_gateway_id": "LocalGatewayId",
            "nat_gateway_id": "NatGatewayId",
            "network_interface_id": "NetworkInterfaceId",
            "transit_gateway_id": "TransitGatewayId",
            "vpc_endpoint_id": "VpcEndpointId",
            "vpc_peering_connection_id": "VpcPeeringConnectionId",
        }
        update_payload = {}
        for new_param, old_param in update_allowed_params.items():
            if (
                locals()[new_param] is not None
                and result["old_state"].get(new_param) != locals()[new_param]
            ):
                update_payload[old_param] = locals()[new_param]
                if destination_cidr_block:
                    update_payload["DestinationCidrBlock"] = destination_cidr_block
                elif destination_ipv6_cidr_block:
                    update_payload[
                        "DestinationIpv6CidrBlock"
                    ] = destination_ipv6_cidr_block
                elif destination_prefix_list_id:
                    update_payload[
                        "DestinationPrefixListId"
                    ] = destination_prefix_list_id

        if update_payload:
            if ctx.get("test", False):
                result["comment"] = hub.tool.aws.comment_utils.would_update_comment(
                    resource_type="aws.ec2.route", name=name
                )
                for new_param, old_param in update_allowed_params.items():
                    plan_state[new_param] = locals()[new_param]
            else:
                update_ret = await hub.exec.boto3.client.ec2.replace_route(
                    ctx, RouteTableId=route_table_id, **update_payload
                )
                if not update_ret.get("result"):
                    result["comment"] = result["comment"] + update_ret["comment"]
                    result["result"] = False
                    return result
                resource_updated = update_ret["result"]
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "route_table_id": route_table_id,
                    "destination_cidr_block": destination_cidr_block,
                    "destination_ipv6_cidr_block": destination_ipv6_cidr_block,
                    "destination_prefix_list_id": destination_prefix_list_id,
                    "vpc_endpoint_id": vpc_endpoint_id,
                    "egress_only_internet_gateway_id": egress_only_internet_gateway_id,
                    "gateway_id": gateway_id,
                    "instance_id": instance_id,
                    "nat_gateway_id": nat_gateway_id,
                    "transit_gateway_id": transit_gateway_id,
                    "local_gateway_id": local_gateway_id,
                    "carrier_gateway_id": carrier_gateway_id,
                    "network_interface_id": network_interface_id,
                    "vpc_peering_connection_id": vpc_peering_connection_id,
                    "core_network_arn": core_network_arn,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.ec2.route", name=name
            )
            return result
        ret = await hub.exec.boto3.client.ec2.create_route(
            ctx,
            DestinationCidrBlock=destination_cidr_block,
            DestinationIpv6CidrBlock=destination_ipv6_cidr_block,
            DestinationPrefixListId=destination_prefix_list_id,
            VpcEndpointId=vpc_endpoint_id,
            EgressOnlyInternetGatewayId=egress_only_internet_gateway_id,
            GatewayId=gateway_id,
            InstanceId=instance_id,
            NatGatewayId=nat_gateway_id,
            TransitGatewayId=transit_gateway_id,
            LocalGatewayId=local_gateway_id,
            CarrierGatewayId=carrier_gateway_id,
            NetworkInterfaceId=network_interface_id,
            RouteTableId=route_table_id,
            VpcPeeringConnectionId=vpc_peering_connection_id,
            CoreNetworkArn=core_network_arn,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.ec2.route", name=name
        )
        resource_id = f"{route_table_id}/{(destination_cidr_block or destination_ipv6_cidr_block or destination_prefix_list_id)}"
    if ctx.get("test", False):
        result["new_state"] = plan_state
    elif (not result["old_state"]) or resource_updated:
        after = await hub.exec.aws.ec2.route.get(
            ctx=ctx, name=name, resource_id=resource_id
        )
        if not after["result"]:
            result["result"] = False
            result["comment"] += after["comment"]
            return result
        result["new_state"] = copy.deepcopy(after["ret"])
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    """Deletes the specified route from the specified route table.

    Args:
        name(str): An Idem name of the resource.
        resource_id(str, Optional): Unique identifier to identify the route in a route table. format of resource id is
            <route_table_id>/<destination_cidr_block or destination_ipv6_cidr_block or destination_prefix_list_id>.
            Idem automatically considers this resource being absent if this field is not specified.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.ec2.route.absent:
                - name: value
                - resource_id: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.route", name=name
        )
        return result

    before = await hub.exec.aws.ec2.route.get(
        ctx=ctx, name=name, resource_id=resource_id
    )
    if not before["result"]:
        result["comment"] = before["comment"]
        result["result"] = False
        return result

    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.route", name=name
        )
    else:
        result["old_state"] = before["ret"]
        if ctx.get("test", False):
            result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
                resource_type="aws.ec2.route", name=name
            )
            return result
        else:
            ret = await hub.exec.boto3.client.ec2.delete_route(
                ctx,
                RouteTableId=result["old_state"]["route_table_id"],
                DestinationCidrBlock=result["old_state"].get("destination_cidr_block"),
                DestinationIpv6CidrBlock=result["old_state"].get(
                    "destination_ipv6_cidr_block"
                ),
                DestinationPrefixListId=result["old_state"].get(
                    "destination_prefix_list_id"
                ),
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = result["comment"] + ret["comment"]
                return result
            result["comment"] = result[
                "comment"
            ] + hub.tool.aws.comment_utils.delete_comment(
                resource_type="aws.ec2.route", name=name
            )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.ec2.route
    """
    result = {}
    ret = await hub.exec.boto3.client.ec2.describe_route_tables(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe route tables {ret['comment']}")
        return result
    for route_table in ret["ret"]["RouteTables"]:
        route_table_id = route_table.get("RouteTableId")
        for route in route_table.get("Routes"):
            resource_key = f"{route_table_id}/{(route.get('DestinationCidrBlock') or route.get('DestinationIpv6CidrBlock') or route.get('DestinationPrefixListId'))}"
            resource_converted = (
                hub.tool.aws.ec2.conversion_utils.convert_raw_route_to_present(
                    raw_resource=route,
                    route_table_id=route_table_id,
                    idem_resource_name=resource_key,
                )
            )
            result[resource_key] = {
                "aws.ec2.route.present": [
                    {parameter_key: parameter_value}
                    for parameter_key, parameter_value in resource_converted.items()
                ]
            }
    return result



"""Exec module for managing KMS Keys."""
from typing import Dict

__func_alias__ = {"list_": "list"}


async def get(
    hub,
    ctx,
    name,
    resource_id,
) -> Dict:
    """Get a KMS Key resource from AWS.

    Args:
        name(str):
            The name of the Idem state.

        resource_id(str):
            Key ARN or alias ARN or alias name
            To specify a KMS key, use its key ID, key ARN, alias name, or alias ARN.
            When using an alias name, prefix it with "alias/". To specify a KMS key in a different Amazon Web Services
            account, you must use the key ARN or alias ARN.

    """
    result = dict(comment=[], ret=None, result=True)
    ret = await hub.exec.boto3.client.kms.describe_key(ctx, KeyId=resource_id)
    if not ret["result"]:
        if "NotFoundException" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.kms.key", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["KeyMetadata"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.kms.key", name=name
            )
        )
        return result

    result["ret"] = await hub.tool.aws.kms.conversion_utils.convert_raw_key_to_present(
        ctx, raw_resource=ret["ret"]["KeyMetadata"]
    )
    return result


async def list_(
    hub,
    ctx,
    name,
) -> Dict:
    """Fetch a list of Kms Keys from AWS.

    Args:
        name(str):
            The name of the Idem state.
    """
    result = dict(comment=[], ret=[], result=True)
    ret = await hub.exec.boto3.client.kms.list_keys(ctx)
    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["Keys"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.kms.key", name=name
            )
        )
        return result
    for key in ret["ret"]["Keys"]:
        key_details = await hub.exec.boto3.client.kms.describe_key(
            ctx, KeyId=key["KeyId"]
        )
        if key_details.get("result") is False:
            hub.log.debug(
                f"Failed to describe key with Id {key.get('KeyId')}: {key_details.get('comment')}"
            )
            continue

        try:
            result["ret"].append(
                await hub.tool.aws.kms.conversion_utils.convert_raw_key_to_present(
                    ctx, raw_resource=key_details["ret"]["KeyMetadata"]
                )
            )
        except Exception as e:
            hub.log.debug(f"Failed to convert key with Id {key.get('KeyId')}: {e}")
            continue
    return result



"""State module for managing EKS Fargate profile."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


TREQ = {
    "present": {
        "require": [
            "aws.eks.cluster.present",
        ],
    },
}
resource_type = "aws.eks.fargate_profile"


async def present(
    hub,
    ctx,
    name: str,
    cluster_name: str,
    pod_execution_role_arn: str,
    resource_id: str = None,
    subnets: List[str] = None,
    selectors: List[
        make_dataclass(
            "FargateProfileSelector",
            [
                ("namespace", str, field(default=None)),
                ("labels", Dict[str, str], field(default=None)),
            ],
        )
    ] = None,
    client_request_token: str = None,
    tags: Dict[str, str] = None,
    timeout: make_dataclass(
        "Timeout",
        [
            (
                "create",
                make_dataclass(
                    "CreateTimeout",
                    [
                        ("delay", int, field(default=10)),
                        ("max_attempts", int, field(default=60)),
                    ],
                ),
                field(default=None),
            ),
            (
                "update",
                make_dataclass(
                    "UpdateTimeout",
                    [
                        ("delay", int, field(default=10)),
                        ("max_attempts", int, field(default=60)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
) -> Dict[str, Any]:
    """Creates an Fargate profile for your Amazon EKS cluster.

    You must have at least one Fargate profile in a cluster to be able to run pods on Fargate. The Fargate profile allows
    an administrator to declare which pods run on Fargate and specify which pods run on which Fargate profile. This
    declaration is done through the profile’s selectors. Each profile can have up to five selectors that contain a
    namespace and labels. A namespace is required for every selector. The label field consists of multiple optional
    key-value pairs. Pods that match the selectors are scheduled on Fargate. If a to-be-scheduled pod matches any of the
    selectors in the Fargate profile, then that pod is run on Fargate. When you create a Fargate profile, you must specify
    a pod execution role to use with the pods that are scheduled with the profile. For more information, see Pod Execution
    Role in the Amazon EKS User Guide. Fargate profiles are immutable. However, you can create a new updated profile to
    replace an existing profile and then delete the original after the updated profile has finished creating. If any
    Fargate profiles in a cluster are in the DELETING status, you must wait for that Fargate profile to finish deleting
    before you can create any other profiles in that cluster. For more information, see Fargate Profile in the Amazon EKS
    User Guide.

    Args:
        name(str):
            An Idem name of the resource

        cluster_name(str):
            The name of the Amazon EKS cluster to apply the Fargate profile to.

        pod_execution_role_arn(str):
            The Amazon Resource Name (ARN) of the pod execution role to use for pods that match the
            selectors in the Fargate profile.

        subnets(list[str], Optional):
            The IDs of subnets to launch your pods into. At this time, pods running on Fargate are not assigned public
            IP addresses, so only private subnets (with no direct route to an Internet Gateway) are accepted for this
            parameter. Defaults to None. Private subnets selected for fargate profile creation must be in different
            availability zones. Defaults to None.

        selectors(list[dict[str, Any]], Optional):
            The selectors to match for pods to use this Fargate profile. Each selector must have an associated namespace.
            Optionally, you can also specify labels for a namespace. You may specify up to five selectors in a Fargate
            profile. Defaults to None.

            * namespace (str, Optional):
                The Kubernetes namespace that the selector should match.

            * labels (dict[str, str], Optional):
                The Kubernetes labels that the selector should match. A pod must contain all of the labels that are
                specified in the selector for it to be considered a match.

        client_request_token(str, Optional):
            Unique, case-sensitive identifier that you provide to ensure the idempotency of the request. Defaults to None.

        tags(dict, Optional):
            The metadata to apply to the Fargate profile to assist with categorization and organization.
            Each tag consists of a key and an Optional value. Defaults to None.

        resource_id(str, Optional):
            AWS Fargate Profile name

        timeout(dict, Optional):
            Timeout configuration for operations on fargate_profile

            * create (dict) -- Timeout configuration for creating fargate_profile
                * delay -- The amount of time in seconds to wait between attempts.
                * max_attempts -- The maximum number of attempts to be made.

    Request Syntax:
        .. code-block:: sls

            fargate_resource_id:
              aws.eks.fargate_profile.present:
                - name: 'string'
                - resource_id: 'string'
                - cluster_name: 'string'
                - pod_execution_role_arn: 'string'
                - subnets:
                  - private_subnet_id
                  - private_subnet_id
                - selectors:
                  - labels: {}
                    namespace: 'string'
                - tags:
                  - 'string': 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            fargate01-idem-cluster01:
              aws.eks.fargate_profile.present:
                - name: fargate01
                - cluster_name: idem-cluster01
                - pod_execution_role_arn: arn:aws:iam::537227425989:role/eksfargateprofile-role
                - subnets:
                  - subnet-006a41c410b9485dd
                  - subnet-01368936bb540ba1e
                - selectors:
                  - labels: {}
                    namespace: default
                - tags:
                    fargate01: 'created via code'
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = await hub.exec.boto3.client.eks.describe_fargate_profile(
        ctx,
        clusterName=cluster_name,
        fargateProfileName=resource_id if resource_id else name,
    )

    # existing resource
    if before and before["ret"]["fargateProfile"]:
        result[
            "old_state"
        ] = hub.tool.aws.eks.conversion_utils.convert_raw_fargate_profile_to_present(
            raw_resource=before["ret"]["fargateProfile"]
        )
        result["new_state"] = copy.deepcopy(result["old_state"])
        result["comment"] = (f"aws.eks.fargate_profile '{name}' already exists",)
        # fargate_profile can only be updated for tags after its creation, but tag update apis are currently
        # supported for eks clusters and nodegroups only. So no update operations are performed
        return result
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "cluster_name": cluster_name,
                    "resource_id": resource_id,
                    "client_request_token": client_request_token,
                    "pod_execution_role_arn": pod_execution_role_arn,
                    "tags": tags,
                    "subnets": subnets,
                    "selectors": selectors,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type=resource_type, name=name
            )
            return result
        # create a new resource
        ret = await hub.exec.boto3.client.eks.create_fargate_profile(
            ctx,
            fargateProfileName=name,
            clusterName=cluster_name,
            podExecutionRoleArn=pod_execution_role_arn,
            subnets=subnets,
            selectors=selectors,
            clientRequestToken=client_request_token,
            tags=tags,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result

        # create waiter config
        waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
            default_delay=10,
            default_max_attempts=60,
            timeout_config=timeout.get("create") if timeout else None,
        )
        hub.log.debug(f"Waiting on creating aws.eks.fargate_profile '{name}'")
        try:
            # wait till fargate_profile is active,
            await hub.tool.boto3.client.wait(
                ctx,
                "eks",
                "fargate_profile_active",
                None,
                clusterName=cluster_name,
                fargateProfileName=name,
                WaiterConfig=waiter_config,
            )
        except Exception as e:
            result["comment"] = result["comment"] + (str(e),)
            result["result"] = False

        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type=resource_type, name=name
        )

    # Get the newly created resource
    result[
        "new_state"
    ] = hub.tool.aws.eks.conversion_utils.convert_raw_fargate_profile_to_present(
        raw_resource=ret["ret"]["fargateProfile"]
    )
    return result


async def absent(
    hub,
    ctx,
    name: str,
    cluster_name: str,
    resource_id: str = None,
    timeout: Dict = None,
) -> Dict[str, Any]:
    """Deletes an Fargate profile.

    When you delete a Fargate profile, any pods running on Fargate that were created with the profile are deleted. If
    those pods match another Fargate profile, then they are scheduled on Fargate with that profile. If they no longer
    match any Fargate profiles, then they are not scheduled on Fargate and they may remain in a pending state. Only one
    Fargate profile in a cluster can be in the DELETING status at a time. You must wait for a Fargate profile to finish
    deleting before you can delete any other profiles in that cluster.

    Args:
        name(str):
            An Idem name of the resource

        cluster_name(str):
            The name of the Amazon EKS cluster associated with the Fargate profile to delete.

        resource_id(str, Optional):
            AWS Fargate profile name. Idem considers name of fargate profile if
         this field is not specified.

        timeout(dict, Optional):
            Timeout configuration for operations on fargate_profile

            * delete (dict) -- Timeout configuration for deleting fargate_profile
                * delay -- The amount of time in seconds to wait between attempts.
                * max_attempts -- The maximum number of attempts to be made

    Request Syntax:
        .. code-block:: sls

            fargate_resource_id:
              aws.eks.fargate_profile.absent:
                - name: 'string'
                - cluster_name: 'string'
    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            fargate01-idem-cluster01:
              aws.eks.fargate_profile.absent:
                - resource_id: fargate01
                - name: fargate01
                - cluster_name: idem-cluster01
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = await hub.exec.boto3.client.eks.describe_fargate_profile(
        ctx,
        clusterName=cluster_name,
        fargateProfileName=resource_id if resource_id else name,
    )
    if before and before["ret"]["fargateProfile"]:
        result[
            "old_state"
        ] = hub.tool.aws.eks.conversion_utils.convert_raw_fargate_profile_to_present(
            raw_resource=before["ret"]["fargateProfile"]
        )

    if not before:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type=resource_type, name=name
        )
    else:
        if ctx.get("test", False):
            result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
                resource_type=resource_type, name=name
            )
        else:
            ret = await hub.exec.boto3.client.eks.delete_fargate_profile(
                ctx,
                clusterName=cluster_name,
                fargateProfileName=resource_id if resource_id else name,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result

            # wait till fargate_profile has been deleted
            waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
                default_delay=30,
                default_max_attempts=60,
                timeout_config=timeout.get("delete") if timeout else None,
            )
            hub.log.debug(f"Waiting on deleting aws.eks.fargate_profile '{name}'")
            try:
                await hub.tool.boto3.client.wait(
                    ctx,
                    "eks",
                    "fargate_profile_deleted",
                    None,
                    fargateProfileName=resource_id,
                    clusterName=cluster_name,
                    WaiterConfig=waiter_config,
                )
            except Exception as e:
                result["comment"] = result["comment"] + (str(e),)
                result["result"] = False
            result["comment"] = hub.tool.aws.comment_utils.delete_comment(
                resource_type=resource_type, name=name
            )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """
    Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Lists the Fargate profiles associated with the specified cluster in your Amazon Web Services account in the
    specified Region.

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:
        .. code-block:: bash

            $ idem describe aws.eks.fargate_profile

    """

    result = {}
    # List all clusters
    cluster_ret = await hub.exec.boto3.client.eks.list_clusters(ctx)
    if not cluster_ret["result"]:
        hub.log.debug(f"Could not list clusters {cluster_ret['comment']}")
        return {}

    fargate_profile_list = []
    # Get fargate profiles in each cluster
    for clusterName in cluster_ret["ret"]["clusters"]:
        fargate_profiles_ret = await hub.exec.boto3.client.eks.list_fargate_profiles(
            ctx, clusterName=clusterName
        )
        if not fargate_profiles_ret["result"]:
            hub.log.info(
                f"No fargate profiles could be listed for cluster {clusterName}. Fetching the next cluster."
            )
            continue

        # maintain a list of clusterName and fargateProfileName for future use
        for fargate_profile_name in fargate_profiles_ret["ret"]["fargateProfileNames"]:
            fargate_profile_list.append(
                {
                    "clusterName": clusterName,
                    "fargateProfileName": fargate_profile_name,
                }
            )
    # Describe each fargate profile
    for fargate_profile in fargate_profile_list:
        clusterName = fargate_profile["clusterName"]
        fargateProfileName = fargate_profile["fargateProfileName"]
        describe_ret = await hub.exec.boto3.client.eks.describe_fargate_profile(
            ctx, clusterName=clusterName, fargateProfileName=fargateProfileName
        )

        if not describe_ret["result"]:
            hub.log.info(
                f"Describe command failed for resource {fargateProfileName} under cluster {clusterName}"
            )
            continue

        fargate_profile = describe_ret["ret"]["fargateProfile"]
        translated_resource = (
            hub.tool.aws.eks.conversion_utils.convert_raw_fargate_profile_to_present(
                raw_resource=fargate_profile
            )
        )
        # state_id is combination of fargateProfileName and clusterName to maintain integrity as more than one
        # cluster can have same fargate profile names
        state_id = (
            fargate_profile.get("fargateProfileName")
            + "-"
            + fargate_profile.get("clusterName")
        )
        result[state_id] = {
            "aws.eks.fargate_profile.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in translated_resource.items()
            ]
        }

    return result



from typing import Dict
from typing import List

__func_alias__ = {"list_": "list"}


async def get(hub, ctx, name, resource_id: str = None, filters: List = None):
    """
    Get a single route table from AWS. If more than one resource is found, the first resource returned from AWS will be used.

    The function returns None when no resource is found.

    Args:
        name(str):
            An Idem state name.

        resource_id(str, Optional):
            AWS route table ID to identify the resource.

        filters(list[dict[str, Any]], Optional):
            One or more filters: for example, tag :<key>, tag-key. A complete list of filters can be found at
            https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_route_tables
    """
    result = dict(comment=[], ret=None, result=True)
    ret = await hub.tool.aws.ec2.route_table.search_raw(
        ctx=ctx, resource_id=resource_id, filters=filters
    )
    if not ret["result"]:
        if "InvalidRouteTableID.NotFound" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.ec2.route_table", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["RouteTables"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.ec2.route_table", name=name
            )
        )
        return result

    resource = ret["ret"]["RouteTables"][0]
    if len(ret["ret"]["RouteTables"]) > 1:
        result["comment"].append(
            hub.tool.aws.comment_utils.find_more_than_one(
                resource_type="aws.ec2.route_table",
                resource_id=resource.get("RouteTableId"),
            )
        )
    result[
        "ret"
    ] = hub.tool.aws.ec2.conversion_utils.convert_raw_route_table_to_present(
        raw_resource=resource, idem_resource_name=name
    )

    return result


async def list_(hub, ctx, name: str = None, filters: List = None) -> Dict:
    """
    Fetch a list of route table from AWS. The function returns empty list when no resource is found.

    Args:
        name(str, Optional):
            An Idem state name.

        filters(list[dict[str, Any]], Optional):
            One or more filters: for example, tag :<key>, tag-key. A complete list of filters can be found at
            https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_route_tables
    """
    result = dict(comment=[], ret=[], result=True)
    ret = await hub.tool.aws.ec2.route_table.search_raw(ctx=ctx, filters=filters)
    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["RouteTables"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.ec2.route_table", name=name
            )
        )
        return result
    for resource in ret["ret"]["RouteTables"]:
        resource_id = resource.get("RouteTableId")
        result["ret"].append(
            hub.tool.aws.ec2.conversion_utils.convert_raw_route_table_to_present(
                raw_resource=resource, idem_resource_name=resource_id
            )
        )
    return result



from typing import Any
from typing import Dict
from typing import List

STOPPED = "stopped"
RUNNING = "running"


async def apply(
    hub,
    ctx,
    resource,
    *,
    old_value: Dict[str, Any],
    new_value: Dict[str, Any],
    comments: List[str],
) -> bool:
    """
    Compare the kwargs of the present function to the presentized attributes of the instance
    """
    result = True

    current_state = old_value
    desired_state = new_value

    # Get the current_state/desired state for running
    initially_running = current_state.get("running")

    # Determine the groups that attributes need to be updated with
    grouping = hub.tool.aws.ec2.instance.update.init.group(
        current_state=current_state, desired_state=desired_state, comments=comments
    )
    if not grouping:
        return False

    # Start with the attributes that work with the instance in the current running state
    # This will minimize number of shutdown/restarts as we update the instance
    if initially_running:
        first_group = grouping["needs_running"]
        second_group = grouping["needs_stopped"]
    else:
        first_group = grouping["needs_stopped"]
        second_group = grouping["needs_running"]

    # Add the attributes that don't care about the current state to the first group
    first_group.extend(grouping["needs_any"])

    # Run the first group
    result &= await hub.tool.aws.ec2.instance.update.init.apply_group(
        ctx,
        resource,
        first_group,
        update_groups=grouping["groups"],
        old_value=old_value,
        new_value=new_value,
        comments=comments,
    )

    # Stop now if we were unable to update the first group.  No need to change the instance's state
    if not result:
        return False

    # Change the state of the instance to the opposite of the starting state
    toggled_running_state = False
    if second_group:
        toggled_running_state = True
        if initially_running:
            ret = await hub.exec.aws.ec2.instance.stop(ctx, instance_id=resource.id)
            result &= ret.result
            if ret.comment:
                comments.append(ret.comment)
        else:
            ret = await hub.exec.aws.ec2.instance.start(ctx, instance_id=resource.id)
            result &= ret.result
            if ret.comment:
                comments.append(ret.comment)

    # Stop now if unable to toggle the running state
    if not result:
        return False

    try:
        # Return with the second group's result
        result &= await hub.tool.aws.ec2.instance.update.init.apply_group(
            ctx,
            resource,
            second_group,
            update_groups=grouping["groups"],
            old_value=old_value,
            new_value=new_value,
            comments=comments,
        )
    # If anything in the second group fails, revert the running state
    finally:
        # The running state is modified here
        # If the new state requires a change in the instance, then leave the state alone
        if toggled_running_state:
            if initially_running or (new_value.get(RUNNING) is True):
                ret = await hub.exec.aws.ec2.instance.start(
                    ctx, instance_id=resource.id
                )
                result &= ret.result
                if ret.comment:
                    comments.append(ret.comment)
            elif (not initially_running) or (new_value.get(RUNNING) is False):
                ret = await hub.exec.aws.ec2.instance.stop(ctx, instance_id=resource.id)
                result &= ret.result
                if ret.comment:
                    comments.append(ret.comment)

    return result


def group(
    hub,
    current_state: Dict[str, Any],
    desired_state: Dict[str, Any],
    comments: List[str],
) -> Dict[str, Any]:
    """
    Group attributes by the state the instance needs to be in for the change
    """
    update_needs_stopped = []
    update_needs_running = []
    update_needs_any = []
    update_group: Dict[str, List[str]] = {}

    # Group attributes together to determine how to do the update
    for attribute, new_attr_value in desired_state.items():
        if desired_state.get(attribute) is None:
            # There is no new value for the attribute
            continue
        if desired_state.get(attribute) == current_state.get(attribute):
            # This specific attribute is in the proper state
            continue

        # Check if this attribute is part of an update group
        # Such as updating affinity and tenancy together as "placement"
        for mod in hub.tool.aws.ec2.instance.update:
            group_ = getattr(mod, "__update_group__", [])
            if attribute in group_:
                attribute_update_group = mod.__name__
                if attribute_update_group not in update_group:
                    update_group[attribute_update_group] = []
                update_group[attribute_update_group].append(attribute)

                # Now treat the update group as the mod and attribute
                attribute = attribute_update_group
                break

        if attribute not in hub.tool.aws.ec2.instance.update:
            comments.append(f"No plugin available to modify attribute {attribute}")
            continue

        # Determine which state this group needs to be in
        mod = hub.tool.aws.ec2.instance.update[attribute]

        # Determine what state the VM needs to be in to update this attribute
        if getattr(mod, "__update_state__", None) == STOPPED:
            # This attribute can only be modified while stopped
            update_needs_stopped.append(attribute)
        elif getattr(mod, "__update_state__", None) == RUNNING:
            # This attribute can only be modified while running
            update_needs_running.append(attribute)
        else:
            # This attribute has no extra needs, add it to the first update group
            update_needs_any.append(attribute)

    return {
        "needs_stopped": update_needs_stopped,
        "needs_running": update_needs_running,
        "needs_any": update_needs_any,
        "groups": update_group,
    }


async def apply_group(
    hub,
    ctx,
    resource,
    current_group: List[str],
    *,
    update_groups: Dict[str, List[str]],
    old_value: Dict[str, Any],
    new_value: Dict[str, Any],
    comments: List[str],
) -> bool:
    """
    Apply updates to an instance based on the current group
    """
    result = True
    for attribute in current_group:
        if ctx.test:
            comments += [f"Would update aws.ec2.instance: {attribute}"]
            continue

        update_group = None
        for g in update_groups:
            if attribute in g:
                update_group = g
                break

        if update_group:
            # pass all the attributes together
            result &= await hub.tool.aws.ec2.instance.update[attribute].apply(
                ctx,
                resource,
                old_value={
                    attr: val
                    for attr, val in old_value.items()
                    if attr in update_groups[update_group]
                },
                new_value={
                    attr: val
                    for attr, val in new_value.items()
                    if attr in update_groups[update_group]
                },
                comments=comments,
            )
            if not result:
                comments.append(
                    f"Unable to update aws.ec2.instance attribute: {attribute}"
                )
                break
            continue

        # Call the appropriate tool to update each parameter that needs updating
        result &= await hub.tool.aws.ec2.instance.update[attribute].apply(
            ctx,
            resource,
            old_value=old_value.get(attribute),
            new_value=new_value.get(attribute),
            comments=comments,
        )
        if not result:
            comments.append(f"Unable to update aws.ec2.instance attribute: {attribute}")
            break
    return result



"""Get and List methods for AWS SES Event destination."""
from typing import Any
from typing import Dict

__func_alias__ = {"list_": "list"}


async def get(
    hub, ctx, configuration_set_name: str, resource_id: str = None
) -> Dict[str, Any]:
    """Retrieve event destinations that are associated with a configuration set.

    Args:
        configuration_set_name(str):
            The name of the configuration set.
        resource_id(str, Optional):
            A name that identifies the event destination within the configuration set.

    Returns:
        .. code-block:: bash

            {"result": True|False, "comment": ("A tuple",), "ret": None|Dict}


    Examples:
        Calling this exec module function from the cli

        .. code-block:: bash

            idem exec aws.sesv2.event_destination.get configuration_set_name=config_set_name resource_id=event_destination_name


        Calling this exec module function from within a state module

        .. code-block:: python

            async def state_function(hub, ctx, name, resource_id, configuration_set_name, **kwargs):
                ret = await hub.exec.aws.sesv2.event_destination.get(ctx=ctx,
                                                                     resource_id=resource_id,
                                                                     configuration_set_name=configuration_set_name)
    """
    ret = {"result": True, "comment": [], "ret": {}}
    config_dest = (
        await hub.exec.boto3.client.sesv2.get_configuration_set_event_destinations(
            ctx, ConfigurationSetName=configuration_set_name
        )
    )
    if not config_dest.get("result"):
        ret["comment"] = config_dest["comment"]
        ret["result"] = False
        return ret
    for d in config_dest.get("ret", {}).get("EventDestinations", []):
        if resource_id and d["Name"] != resource_id:
            continue
        ret[
            "ret"
        ] = hub.tool.aws.sesv2.conversion_utils.convert_raw_event_destination_to_present(
            configuration_set_name=configuration_set_name,
            raw_resource=d,
            name=resource_id,
        )
        break
    if not ret["ret"]:
        ret["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.sesv2.event_destination", name=resource_id
            )
        )
    return ret


async def list_(hub, ctx) -> Dict[str, Any]:
    """Retrieve a list of event destinations that are associated with a configuration set.

    Returns:
        .. code-block:: bash

            {"result": True|False, "comment": ("A tuple",), "ret": None|Dict}


    Examples:
        Calling this exec module function from the cli

        .. code-block:: bash

            idem exec aws.sesv2.event_destination.list


        Calling this exec module function from within a state module

        .. code-block:: python

            async def state_function(hub, ctx, name, resource_id, configuration_set_name, **kwargs):
                ret = await hub.exec.aws.sesv2.event_destination.list(ctx=ctx)
    """
    result = {"result": True, "comment": [], "ret": {}}
    ret = await hub.exec.boto3.client.sesv2.list_configuration_sets(ctx)

    if not ret["result"]:
        hub.log.error(f"Could not list configuration sets {ret['comment']}")
        result["comment"] = ret["comment"]
        result["result"] = False
        return result

    for name in ret["ret"].get("ConfigurationSets", []):
        resp = (
            await hub.exec.boto3.client.sesv2.get_configuration_set_event_destinations(
                ctx, ConfigurationSetName=name
            )
        )
        for e in resp["ret"].get("EventDestinations", []):
            new_event_destination = [
                {"name": e["Name"]},
                {"resource_id": e["Name"]},
                {"configuration_set_name": name},
                {"event_destination": e.pop("Name") and e.copy()},
            ]
            result["ret"][name] = {
                "aws.sesv2.event_destination.present": new_event_destination
            }
    if not result["ret"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.sesv2.event_destination", name=""
            )
        )
    return result



"""State module for managing Amazon Elasticsearch Service Domain"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    domain_name: str,
    elastic_search_version: str,
    perform_check_only: bool = None,
    elastic_search_cluster_config: make_dataclass(
        """Configuration to specify the instance type and number of instances in the domain cluster."""
        "ElasticsearchClusterConfig",
        [
            ("InstanceType", str, field(default=None)),
            ("InstanceCount", str, field(default=None)),
            ("DedicatedMasterEnabled", bool, field(default=None)),
            ("ZoneAwarenessEnabled", bool, field(default=None)),
            (
                "ZoneAwarenessConfig",
                make_dataclass(
                    "ZoneAwarenessConfig",
                    [("AvailabilityZoneCount", int, field(default=None))],
                ),
                field(default=None),
            ),
            ("DedicatedMasterType", str, field(default=None)),
            ("DedicatedMasterCount", int, field(default=None)),
            ("WarmEnabled", bool, field(default=None)),
            ("WarmType", str, field(default=None)),
            ("WarmCount", int, field(default=None)),
            (
                "ColdStorageOptions",
                make_dataclass(
                    "ColdStorageOptions",
                    [("Enabled", bool)],
                ),
                field(default=None),
            ),
            ("OnFailure", Dict, field(default=None)),
        ],
    ) = None,
    ebs_options: make_dataclass(
        """Options to enable, disable and specify the type and size of EBS storage volumes."""
        "EBSOptions",
        [
            ("EBSEnabled", bool, field(default=None)),
            ("VolumeType", str, field(default=None)),
            ("VolumeSize", int, field(default=None)),
            ("Iops", int, field(default=None)),
            ("Throughput", int, field(default=None)),
        ],
    ) = None,
    access_policies: str = None,
    snapshot_options: make_dataclass(
        """Option to set time, in UTC format, of the daily automated snapshot. Default value is 0 hours."""
        "SnapshotOptions",
        [
            ("AutomatedSnapshotStartHour", int, field(default=None)),
        ],
    ) = None,
    vpc_options: make_dataclass(
        """Options to specify the subnets and security groups for VPC endpoint"""
        "VPCOptions",
        [
            ("SubnetIds", List[str], field(default=None)),
            ("SecurityGroupIds", List[str], field(default=None)),
        ],
    ) = None,
    cognito_options: make_dataclass(
        """Options to specify the Cognito user and identity pools for Kibana authentication."""
        "CognitoOptions",
        [
            ("Enabled", bool, field(default=None)),
            ("UserPoolId", str, field(default=None)),
            ("IdentityPoolId", str, field(default=None)),
            ("RoleArn", str, field(default=None)),
        ],
    ) = None,
    encryption_at_rest_options: make_dataclass(
        """Encryption At Rest Options.""" "EncryptionAtRestOptions",
        [
            ("Enabled", bool, field(default=None)),
            ("KmsKeyId", str, field(default=None)),
        ],
    ) = None,
    node_to_node_encryption_options: make_dataclass(
        """Specifies the NodeToNodeEncryptionOptions.""" "NodeToNodeEncryptionOptions",
        [
            ("Enabled", bool, field(default=None)),
        ],
    ) = None,
    advanced_options: Dict[str, str] = None,
    log_publishing_options: Dict[str, Any] = None,
    domain_endpoint_options: make_dataclass(
        """Options to specify configuration that will be applied to the domain endpoint."""
        "DomainEndpointOptions",
        [
            ("EnforceHTTPS", bool, field(default=None)),
            ("TLSSecurityPolicy", str, field(default=None)),
            ("CustomEndpointEnabled", bool, field(default=None)),
            ("CustomEndpoint", str, field(default=None)),
            ("CustomEndpointCertificateArn", str, field(default=None)),
        ],
    ) = None,
    advanced_security_options: make_dataclass(
        """Specify advanced security options.""" "AdvancedSecurityOptions",
        [
            ("Enabled", bool, field(default=None)),
            ("InternalUserDatabaseEnabled", bool, field(default=None)),
            (
                "MasterUserOptions",
                make_dataclass(
                    "MasterUserOptions",
                    [
                        ("MasterUserARN", str, field(default=None)),
                        ("MasterUserName", str, field(default=None)),
                        ("MasterUserPassword", str, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
            (
                "SAMLOptions",
                make_dataclass(
                    "SAMLOptions",
                    [
                        ("Enabled", bool, field(default=None)),
                        (
                            "Idp",
                            make_dataclass(
                                "Idp",
                                [
                                    ("MetadataContent", str),
                                    ("EntityId", str),
                                ],
                            ),
                            field(default=None),
                        ),
                        ("MasterUserName", str, field(default=None)),
                        ("MasterBackendRole", str, field(default=None)),
                        ("SubjectKey", str, field(default=None)),
                        ("RolesKey", str, field(default=None)),
                        ("SessionTimeoutMinutes", int, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
            ("AnonymousAuthEnabled", bool, field(default=None)),
        ],
    ) = None,
    auto_tune_options: make_dataclass(
        """Specifies Auto-Tune options.""" "AutoTuneOptions",
        [
            ("DesiredState", str, field(default=None)),
            (
                "MaintenanceSchedules",
                List[
                    make_dataclass(
                        "MaintenanceSchedules",
                        [
                            ("StartAt", str, field(default=None)),
                            (
                                "Duration",
                                make_dataclass(
                                    "Duration",
                                    [
                                        ("Value", int, field(default=None)),
                                        ("Unit", str, field(default=None)),
                                    ],
                                ),
                                field(default=None),
                            ),
                            ("CronExpressionForRecurrence", str, field(default=None)),
                        ],
                    )
                ],
                field(default=None),
            ),
            ("RollbackOnDisable", str, field(default=None)),
        ],
    ) = None,
    dry_run: bool = None,
    tags: Dict[str, str] = None,
    timeout: make_dataclass(
        "Timeout",
        [
            (
                "create",
                make_dataclass(
                    "CreateTimeout",
                    [
                        ("delay", int, field(default=30)),
                        ("max_attempts", int, field(default=60)),
                    ],
                ),
                field(default=None),
            ),
            (
                "update",
                make_dataclass(
                    "UpdateTimeout",
                    [
                        ("delay", int, field(default=30)),
                        ("max_attempts", int, field(default=60)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    """Creates a new Elasticsearch domain.

    Args:
        name(str): An Idem name of the resource.
        domain_name(str):
            The name of the Elasticsearch domain that you are creating. Domain names are unique across the domains owned
            by an account within an AWS region. Domain names must start with a lowercase letter and can contain the
            following characters: a-z (lowercase), 0-9, and - (hyphen).
        elastic_search_version(str):
            String of format X.Y to specify version for the Elasticsearch domain
            eg. "1.5" or "2.3".
        perform_check_only(bool):
            This flag, when set to True, indicates that an Upgrade Eligibility Check needs to be performed. This will not actually perform the Upgrade.

                .. note::
                    This property can only be applied when `elastic_search_version` is upgraded
        elastic_search_cluster_config(dict, Optional):
            Configuration options for an Elasticsearch domain. Specifies the instance type and number of instances in the domain cluster.
                * InstanceType(str, Optional): The instance type for an Elasticsearch cluster. UltraWarm instance types are not supported for data instances.
                * InstanceCount(int, Optional): The number of instances in the specified domain cluster.
                * DedicatedMasterEnabled(bool, Optional): A boolean value to indicate whether a dedicated master node is enabled.
                * ZoneAwarenessEnabled(bool, Optional): A boolean value to indicate whether zone awareness is enabled.
                * ZoneAwarenessConfig(dict, Optional):
                    Specifies the zone awareness configuration for a domain when zone awareness is enabled.
                        * AvailabilityZoneCount(int, Optional): An integer value to indicate the number of availability zones for a domain when zone awareness is enabled. This should be equal to number of subnets if VPC endpoints is enabled
                * DedicatedMasterType(str, Optional): The instance type for a dedicated master node.
                * DedicatedMasterCount(int, Optional): Total number of dedicated master nodes, active and on standby, for the cluster.
                * WarmEnabled(bool, Optional): True to enable warm storage.
                * WarmType(str, Optional): The instance type for the Elasticsearch cluster's warm nodes.
                * WarmCount(int, Optional): The number of warm nodes in the cluster.
                * ColdStorageOptions(dict, Optional):
                    Specifies the ColdStorageOptions config for Elasticsearch Domain
                        * Enabled(bool): Enable cold storage option. Accepted values true or false
        ebs_options(dict, Optional):
            Options to enable, disable and specify the type and size of EBS storage volumes.
                * EBSEnabled(bool, Optional): Specifies whether EBS-based storage is enabled.
                * VolumeType(str, Optional): Specifies the volume type for EBS-based storage.
                * VolumeSize(int, Optional): Integer to specify the size of an EBS volume.
                * Iops(int, Optional): Specifies the IOPS for Provisioned IOPS And GP3 EBS volume (SSD).
                * Throughput(int, Optional): Specifies the Throughput for GP3 EBS volume (SSD).
        access_policies(str, Optional): IAM access policy as a JSON-formatted string.
        snapshot_options(dict, Optional):
            Option to set time, in UTC format, of the daily automated snapshot. Default value is 0 hours.
                * AutomatedSnapshotStartHour(int, Optional): Specifies the time, in UTC format, when the service takes a daily automated snapshot of the specified Elasticsearch domain. Default value is 0 hours.
        vpc_options(dict, Optional):
            Options to specify the subnets and security groups for VPC endpoint.
                * SubnetIds (list[str], Optional): Specifies the subnets for VPC endpoint.
                * SecurityGroupIds (list[str], Optional): Specifies the security groups for VPC endpoint.
        cognito_options(dict, Optional):
            Options to specify the Cognito user and identity pools for Kibana authentication.
                * Enabled(bool, Optional): Specifies the option to enable Cognito for Kibana authentication.
                * UserPoolId(str, Optional): Specifies the Cognito user pool ID for Kibana authentication.
                * IdentityPoolId(str, Optional): Specifies the Cognito identity pool ID for Kibana authentication.
                * RoleArn(str, Optional): Specifies the role ARN that provides Elasticsearch permissions for accessing Cognito resources.
        encryption_at_rest_options(dict, Optional):
            Specifies the Encryption At Rest Options.
                * Enabled(bool, Optional): Specifies the option to enable Encryption At Rest.
                * KmsKeyId(str, Optional): Specifies the KMS Key ID for Encryption At Rest options.
        node_to_node_encryption_options(dict, Optional):
            Specifies the NodeToNodeEncryptionOptions.
                * Enabled(bool, Optional): Specify true to enable node-to-node encryption.
        advanced_options(dict, Optional): Option to allow references to indices in an HTTP request body. Must be false when configuring access to individual sub-resources. By default, the value is true.
        log_publishing_options(dict, Optional):
            Map of LogType and LogPublishingOption , each containing options to publish a given type of Elasticsearch log.
            Type of Log File, it can be one of the following:
            * INDEX_SLOW_LOGS: Index slow logs contain insert requests that took more time than configured index query log threshold to execute.
            * SEARCH_SLOW_LOGS: Search slow logs contain search queries that took more time than configured search query log threshold to execute.
            * ES_APPLICATION_LOGS: Elasticsearch application logs contain information about errors and warnings raised during the operation of the service and can be useful for troubleshooting.
            * AUDIT_LOGS: Audit logs contain records of user requests for access from the domain.
            Log Publishing option that is set for given domain. Attributes and their details:
            * CloudWatchLogsLogGroupArn: ARN of the Cloudwatch log group to which log needs to be published.
            * Enabled: Specifies whether given log publishing option is enabled or not.
        domain_endpoint_options(dict, Optional):
            Options to specify configuration that will be applied to the domain endpoint.
                * EnforceHTTPS(bool, Optional): Specify if only HTTPS endpoint should be enabled for the Elasticsearch domain.
                * TLSSecurityPolicy(str, Optional):
                    Specify the TLS security policy that needs to be applied to the HTTPS endpoint of Elasticsearch domain. It can be one of the following values:
                        Policy-Min-TLS-1-0-2019-07: TLS security policy which supports TLSv1.0 and higher.
                        Policy-Min-TLS-1-2-2019-07: TLS security policy which supports only TLSv1.2
                * CustomEndpointEnabled(bool, Optional): Specify if custom endpoint should be enabled for the Elasticsearch domain.
                * CustomEndpoint(str, Optional): Specify the fully qualified domain for your custom endpoint.
                * CustomEndpointCertificateArn(str, Optional): Specify ACM certificate ARN for your custom endpoint..
        advanced_security_options(dict, Optional):
            Specifies advanced security options.
                * Enabled(bool, Optional): True if advanced security is enabled.
                * InternalUserDatabaseEnabled(bool, Optional): True if the internal user database is enabled.
                * MasterUserOptions(dict, Optional):
                    Credentials for the master user: username and password, ARN, or both.
                        * MasterUserARN(str, Optional): ARN for the master user (if IAM is enabled).
                        * MasterUserName(str, Optional): The master user's username, which is stored in the Amazon Elasticsearch Service domain's internal database.
                        * MasterUserPassword(str, Optional): The master user's password, which is stored in the Amazon Elasticsearch Service domain's internal database.
                * SAMLOptions(dict, Optional):
                    Specifies the SAML application configuration for the domain.
                        * Enabled(bool, Optional): True if SAML is enabled.
                        * Idp(dict, Optional):
                            Specifies the SAML Identity Provider's information.
                                * MetadataContent(str): The Metadata of the SAML application in xml format.
                                * EntityId(str): The unique Entity ID of the application in SAML Identity Provider.
                        * MasterUserName(str, Optional): The SAML master username, which is stored in the Amazon Elasticsearch Service domain's internal database.
                        * MasterBackendRole(str, Optional): The backend role to which the SAML master user is mapped to.
                        * SubjectKey(str, Optional): The key to use for matching the SAML Subject attribute.
                        * MasterUserName(str, Optional): The SAML master username, which is stored in the Amazon Elasticsearch Service domain's internal database.
                        * MasterBackendRole(str, Optional): The backend role to which the SAML master user is mapped to.
                        * SubjectKey(str, Optional): The key to use for matching the SAML Subject attribute.
                        * RolesKey(str, Optional): The key to use for matching the SAML Roles attribute
                        * SessionTimeoutMinutes(int, Optional): The duration, in minutes, after which a user session becomes inactive. Acceptable values are between 1 and 1440, and the default value is 60.
                * AnonymousAuthEnabled(bool, Optional): True if Anonymous auth is enabled. Anonymous auth can be enabled only when AdvancedSecurity is enabled on existing domains.
        auto_tune_options(dict, Optional):
            Specifies Auto-Tune options.
                * DesiredState(str, Optional): Specifies the Auto-Tune desired state. Valid values are ENABLED, DISABLED.
                * RollbackOnDisable(str, Optional):
                    Specifies the rollback state while disabling Auto-Tune for the domain. Valid values are NO_ROLLBACK, DEFAULT_ROLLBACK.

                    .. note::
                        This property is only applied on Update of the resource
                * MaintenanceSchedules(list[dict[str, Any]], Optional):
                    Specifies list of maintenance schedules.
                        * StartAt(datetime, Optional): Specifies timestamp at which Auto-Tune maintenance schedule start.
                        * Duration(dict, Optional):
                            Specifies maintenance schedule duration: duration value and duration unit.
                                * Value(int, Optional): Integer to specify the value of a maintenance schedule duration.
                                * Unit(str, Optional): Specifies the unit of a maintenance schedule duration. Valid value is HOURS.
                        * CronExpressionForRecurrence(str, Optional): Specifies cron expression for a recurring maintenance schedule.
        dry_run(bool, Optional):
            This flag, when set to True, specifies whether the update request should return the results of validation checks without actually applying the change. This will not actually perform the Update.

            .. note::
                This property is only applied on Update of the resource
        tags(dict, Optional):
            Dict in the format of {tag-key: tag-value} Tag keys must be unique for the Elasticsearch domain to which they are attached.
            Tag values can be null and do not have to be unique in a tag set. For example, you can have a key value pair in a tag set of ``project : Trinity`` and ``cost-center : Trinity``
            Defaults to None.
        timeout(dict, Optional):
            Timeout configuration for create/update of AWS Elasticsearch domain
                * create(dict, Optional):
                    Timeout configuration for creating AWS Elasticsearch domain
                        * delay(int, Optional): The amount of time in seconds to wait between attempts. Default value is ``30``.
                        * max_attempts(int, Optional): Customized timeout configuration containing delay and max attempts. Default value is ``60``.
                * update(dict, Optional):
                    Timeout configuration for updating AWS Elasticsearch domain
                        * delay(int, Optional): The amount of time in seconds to wait between attempts. Default value is ``30``.
                        * max_attempts(int, Optional): Customized timeout configuration containing delay and max attempts. Default value is ``60``.
        resource_id(str, Optional): The name of the Elasticsearch domain. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:

         .. code-block:: sls

             resource_is_present:
                  aws.es.elasticsearch_domain.present:
                    - name: value
                    - domain_name: value
                    - elastic_search_version: value

    """
    result = dict(comment=[], old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    tags_updated = False

    if resource_id:
        before = await hub.exec.aws.es.elasticsearch_domain.get(
            ctx, name=name, resource_id=resource_id
        )
        if not (before["result"] and before["ret"]):
            result["comment"] = list(before["comment"])
            result["result"] = False
            return result
        result["old_state"] = copy.deepcopy(before["ret"])
        result["new_state"] = copy.deepcopy(result["old_state"])
        desired_resource_parameters = {
            "domain_name": domain_name,
            "elastic_search_version": elastic_search_version,
            "perform_check_only": perform_check_only,
            "elastic_search_cluster_config": elastic_search_cluster_config,
            "ebs_options": ebs_options,
            "access_policies": access_policies,
            "snapshot_options": snapshot_options,
            "vpc_options": vpc_options,
            "cognito_options": cognito_options,
            "encryption_at_rest_options": encryption_at_rest_options,
            "node_to_node_encryption_options": node_to_node_encryption_options,
            "advanced_options": advanced_options,
            "log_publishing_options": log_publishing_options,
            "domain_endpoint_options": domain_endpoint_options,
            "advanced_security_options": advanced_security_options,
            "auto_tune_options": auto_tune_options,
            "dry_run": dry_run,
        }
        update_elasticsearch_domain_ret = (
            await hub.tool.aws.es.elasticsearch_domain.update_elasticsearch_domain(
                ctx,
                resource_id=resource_id,
                current_state=result["old_state"],
                desired_state=desired_resource_parameters,
                timeout=timeout,
            )
        )
        if not update_elasticsearch_domain_ret["result"]:
            result["result"] = False
            result["comment"] = list(update_elasticsearch_domain_ret["comment"])
            return result
        result["comment"] = list(update_elasticsearch_domain_ret["comment"])

        resource_updated = bool(update_elasticsearch_domain_ret["ret"])

        if tags is not None and tags != result["old_state"].get("tags"):
            # Update tags
            update_tags = await hub.tool.aws.es.tag.update(
                ctx=ctx,
                resource_id=resource_id,
                old_tags=result["old_state"].get("tags"),
                new_tags=tags,
            )
            result["comment"] += list(update_tags["comment"])
            result["result"] = update_tags["result"]
            if update_tags["ret"] is not None:
                tags_updated = True

        if ctx.get("test", False):
            if resource_updated:
                result["new_state"].update(update_elasticsearch_domain_ret["ret"])
            if tags_updated:
                result["new_state"].update({"tags": update_tags["ret"]})
            return result

    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "domain_name": domain_name,
                    "elastic_search_version": elastic_search_version,
                    "elastic_search_cluster_config": elastic_search_cluster_config,
                    "ebs_options": ebs_options,
                    "access_policies": access_policies,
                    "snapshot_options": snapshot_options,
                    "vpc_options": vpc_options,
                    "cognito_options": cognito_options,
                    "encryption_at_rest_options": encryption_at_rest_options,
                    "node_to_node_encryption_options": node_to_node_encryption_options,
                    "advanced_options": advanced_options,
                    "log_publishing_options": log_publishing_options,
                    "domain_endpoint_options": domain_endpoint_options,
                    "advanced_security_options": advanced_security_options,
                    "auto_tune_options": auto_tune_options,
                    "tags": tags,
                },
            )
            result["comment"] = list(
                hub.tool.aws.comment_utils.would_create_comment(
                    resource_type="aws.es.elasticsearch_domain", name=name
                )
            )
            return result

        create_ret = await hub.exec.boto3.client.es.create_elasticsearch_domain(
            ctx,
            DomainName=domain_name,
            ElasticsearchVersion=elastic_search_version,
            ElasticsearchClusterConfig=elastic_search_cluster_config,
            EBSOptions=ebs_options,
            AccessPolicies=access_policies,
            SnapshotOptions=snapshot_options,
            VPCOptions=vpc_options,
            CognitoOptions=cognito_options,
            EncryptionAtRestOptions=encryption_at_rest_options,
            NodeToNodeEncryptionOptions=node_to_node_encryption_options,
            AdvancedOptions=advanced_options,
            LogPublishingOptions=log_publishing_options,
            DomainEndpointOptions=domain_endpoint_options,
            AdvancedSecurityOptions=advanced_security_options,
            AutoTuneOptions=auto_tune_options,
            TagList=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags=tags),
        )
        result["result"] = create_ret["result"]
        if not result["result"]:
            result["comment"] = list(create_ret["comment"])
            return result
        result["comment"] = list(
            hub.tool.aws.comment_utils.create_comment(
                resource_type="aws.es.elasticsearch_domain", name=name
            )
        )
        elasticsearch_domain = create_ret["ret"].get("DomainStatus")
        resource_id = elasticsearch_domain.get("DomainName")

        create_waiter_acceptors = [
            {
                "matcher": "path",
                "expected": True,
                "state": "success",
                "argument": "DomainStatus.Created",
            },
            {
                "matcher": "path",
                "expected": False,
                "state": "retry",
                "argument": "DomainStatus.Created",
            },
        ]

        waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
            default_delay=30,
            default_max_attempts=60,
            timeout_config=timeout.get("create") if timeout else None,
        )

        create_domain_waiter = hub.tool.boto3.custom_waiter.waiter_wrapper(
            name="ElasticsearchDomainCreate",
            operation="DescribeElasticsearchDomain",
            argument=["DomainStatus.Created"],
            acceptors=create_waiter_acceptors,
            client=await hub.tool.boto3.client.get_client(ctx, "es"),
        )
        try:
            await hub.tool.boto3.client.wait(
                ctx,
                "es",
                "ElasticsearchDomainCreate",
                create_domain_waiter,
                DomainName=resource_id,
                WaiterConfig=waiter_config,
            )
        except Exception as e:
            result["comment"] = result["comment"] + (str(e),)
            result["result"] = False
            return result

    if (not before) or resource_updated or tags_updated:
        after = await hub.exec.aws.es.elasticsearch_domain.get(
            ctx, name=name, resource_id=resource_id
        )
        if not (after["result"] and after["ret"]):
            result["result"] = False
            result["comment"] += list(after["comment"])
            return result
        result["new_state"] = copy.deepcopy(after["ret"])

    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    timeout: make_dataclass(
        """Specifies timeout for deletion of domain.""" "Timeout",
        [
            (
                "delete",
                make_dataclass(
                    "DeleteTimeout",
                    [
                        ("delay", int, field(default=30)),
                        ("max_attempts", int, field(default=60)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
) -> Dict[str, Any]:
    """Deletes the specified Elasticsearch domain and all of its data. Once a domain is deleted, it cannot be recovered.

    Args:
        name(str): An Idem name of the resource.
        resource_id(str, Optional): The name of the domain. Idem automatically considers this
            resource being absent if this field is not specified.
        timeout(Dict, Optional):
            Timeout configuration for AWS Elasticsearch domain
                * delete(Dict, Optional):
                    Timeout configuration when deleting an AWS Elasticsearch domain
                        * delay(int, Optional) -- The amount of time in seconds to wait between attempts. Default value is ``30``.
                        * max_attempts(int, Optional) -- Max attempts of waiting for change. Default value is ``60``.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            test-domain-name:
              aws.es.elasticsearch_domain.absent:
                - name: test-domain-name
                - resource_id: test-domain-name
    """
    already_absent_msg = hub.tool.aws.comment_utils.already_absent_comment(
        resource_type="aws.es.elasticsearch_domain", name=name
    )
    result = dict(
        comment=already_absent_msg,
        old_state=None,
        new_state=None,
        name=name,
        result=True,
    )

    if not resource_id:
        return result

    before = await hub.exec.aws.es.elasticsearch_domain.get(
        ctx, name=name, resource_id=resource_id
    )
    if not before["result"]:
        result["comment"] = before["comment"]
        result["result"] = False
        return result

    if before["ret"]:
        result["old_state"] = before["ret"]
        if ctx.get("test", False):
            result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
                resource_type="aws.es.elasticsearch_domain", name=name
            )
        else:
            ret = await hub.exec.boto3.client.es.delete_elasticsearch_domain(
                ctx, DomainName=resource_id
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = hub.tool.aws.comment_utils.delete_comment(
                resource_type="aws.es.elasticsearch_domain", name=name
            )
            delete_waiter_acceptors = [
                {
                    "matcher": "path",
                    "expected": False,
                    "state": "retry",
                    "argument": "DomainStatus.Deleted",
                },
                {
                    "matcher": "path",
                    "expected": True,
                    "state": "retry",
                    "argument": "DomainStatus.Deleted",
                },
                {
                    "matcher": "error",
                    "expected": "ResourceNotFoundException",
                    "state": "success",
                    "argument": "Error.Code",
                },
            ]

            delete_waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
                default_delay=30,
                default_max_attempts=60,
                timeout_config=timeout.get("delete") if timeout else None,
            )

            delete_domain_waiter = hub.tool.boto3.custom_waiter.waiter_wrapper(
                name="ElasticsearchDomainDelete",
                operation="DescribeElasticsearchDomain",
                argument=["DomainStatus.Deleted"],
                acceptors=delete_waiter_acceptors,
                client=await hub.tool.boto3.client.get_client(ctx, "es"),
            )
            try:
                await hub.tool.boto3.client.wait(
                    ctx,
                    "es",
                    "ElasticsearchDomainDelete",
                    delete_domain_waiter,
                    DomainName=resource_id,
                    WaiterConfig=delete_waiter_config,
                )
            except Exception as e:
                result["comment"] = result["comment"] + (str(e),)
                result["result"] = False

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """
    Retrieves a list of domain configuration information of type "Elasticsearch"

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:
        .. code-block:: bash

            $ idem describe aws.es.elasticsearch_domain
    """
    result = {}
    # Returns the domain names of type Elasticsearch. Opensearch domains cannot be managed using this resource state functions
    get_domain_names_ret = await hub.exec.boto3.client.es.list_domain_names(
        ctx, EngineType="Elasticsearch"
    )

    if not get_domain_names_ret["result"]:
        hub.log.debug(f"Could not list domain names {get_domain_names_ret['comment']}")
        return result

    for domain in get_domain_names_ret["ret"]["DomainNames"]:
        domain_name = domain.get("DomainName")
        get_elasticsearch_domain_ret = await hub.exec.aws.es.elasticsearch_domain.get(
            ctx, name=domain_name, resource_id=domain_name
        )

        if not (
            get_elasticsearch_domain_ret["result"]
            and get_elasticsearch_domain_ret["ret"]
        ):
            hub.log.debug(
                f"Could not get domain configuration information for the Elasticsearch domain '{domain_name}': "
                f"{get_elasticsearch_domain_ret['comment']}. Describe will skip this domain and continue."
            )
            continue

        result[domain_name] = {
            "aws.es.elasticsearch_domain.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in get_elasticsearch_domain_ret[
                    "ret"
                ].items()
            ]
        }
    return result



from typing import Any
from typing import Dict


def convert_raw_backup_vault_to_present(
    hub, ctx, raw_resource: Dict[str, Any], vault_name: str
) -> Dict[str, Any]:
    """
    Convert the backup vault response to a common format

    Args:
        raw_resource: Dictionary of backup vault
        vault_name: Name of the backup vault.

    Returns:
        A dictionary of the backup vault
    """
    translated_resource = {}
    raw_resource.pop("ResponseMetadata", None)

    if raw_resource:
        translated_resource["name"] = vault_name
        translated_resource["resource_id"] = vault_name
        translated_resource["backup_vault_arn"] = raw_resource["BackupVaultArn"]
        translated_resource["encryption_key_arn"] = raw_resource["EncryptionKeyArn"]
        translated_resource["number_of_recovery_points"] = raw_resource[
            "NumberOfRecoveryPoints"
        ]
        if raw_resource.get("Tags") is not None:
            translated_resource["tags"] = raw_resource.get("Tags")

    return translated_resource



import copy
from collections import OrderedDict
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List


async def convert_raw_domain_name_to_present(
    hub,
    raw_resource: Dict[str, Any],
    idem_resource_name: str = None,
) -> Dict[str, Any]:
    resource_parameters = OrderedDict(
        {
            "domainName": "domain_name",
            "regionalDomainName": "regional_domain_name",
            "certificateName": "certificate_name",
            "certificateUploadDate": "certificate_upload_date",
            "certificateArn": "certificate_arn",
            "regionalHostedZoneId": "regional_hosted_zone_id",
            "securityPolicy": "security_policy",
            "domainNameStatus": "domain_name_status",
            "endpointConfiguration": "endpoint_configuration",
            "distributionHostedZoneId": "distribution_hosted_zone_id",
            "distributionDomainName": "distribution_domain_name",
            "regionalCertificateArn": "regional_certificate_arn",
            "domainNameStatus": "domain_name_status",
            "domainNameStatusMessage": "domain_name_status_message",
            "securityPolicy": "security_policy",
            "tags": "tags",
            "mutualTlsAuthentication": "mutual_tls_authentication",
            "ownershipVerificationCertificateArn": "ownership_verification_certificate_arn",
        }
    )
    resource_id = raw_resource.get("domainName")
    resource_translated = {
        "resource_id": resource_id,
    }
    if idem_resource_name:
        resource_translated["name"] = idem_resource_name
    for parameter_raw, parameter_present in resource_parameters.items():
        if raw_resource.get(parameter_raw):
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)
    return resource_translated


async def update(
    hub,
    ctx,
    name,
    old_state: Dict[str, Any],
    updatable_parameters: Dict[str, Any],
    timeout: make_dataclass(
        "Timeout",
        [
            (
                "update",
                make_dataclass(
                    "UpdateTimeout",
                    [
                        ("delay", int, field(default=30)),
                        ("max_attempts", int, field(default=60)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
):
    """
    Updates an API Gateway domain_name.

    Args:
        name(str):
            Apigateway domain name
        old_state(dict):
            Previous state of the resource.
        updatable_parameters(dict):
            Parameters from SLS File.
        timeout(dict, Optional): Timeout configuration for updating AWS ApiGateway Domain Name.

    Returns:
        Dict[str, Any]
    """
    result = dict(comment=(), result=True, ret=None)
    patch_operations = []

    new_state = copy.deepcopy(old_state)
    to_be_updated = False
    resource_id = name

    for key, value in updatable_parameters.items():
        if key == "certificate_name":
            key_final = "certificateName"
        elif key == "certificate_arn":
            key_final = "certificateArn"
        elif key == "endpoint_configuration":
            key_final = "endpointConfiguration/types"
        elif key == "mutual_tls_authentication_input":
            key_final = "mutualTlsAuthentication"
        elif key == "regional_certificate_name":
            key_final = "regionalCertificateName"
        elif key == "regional_certificate_arn":
            key_final = "regionalCertificateArn"
        elif key == "security_policy":
            key_final = "securityPolicy"

        if (
            old_state.get(key) is not None
            and value is not None
            and value != old_state.get(key)
        ):
            patch_operations.append(
                {
                    "op": "replace",
                    "path": f"/{key_final}",
                    "value": str(value)
                    if not isinstance(value, List)
                    else ",".join(value),
                }
            )
            to_be_updated = True
        elif old_state.get(key) is None and value is not None:
            patch_operations.append(
                {
                    "op": "add",
                    "path": f"/{key_final}",
                    "value": str(value)
                    if not isinstance(value, List)
                    else ",".join(value),
                }
            )
            new_state[key] = value
            to_be_updated = True
    if ctx.get("test", False):
        if to_be_updated:
            result["comment"] = hub.tool.aws.comment_utils.would_update_comment(
                resource_type="aws.apigateway.domain_name", name=old_state["name"]
            )
            result["ret"] = new_state
        else:
            result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
                resource_type="aws.apigateway.domain_name", name=old_state["name"]
            )
            return result

    if to_be_updated:
        update_ret = await hub.exec.boto3.client.apigateway.update_domain_name(
            ctx, domainName=name, patchOperations=patch_operations
        )
        result["comment"] = hub.tool.aws.comment_utils.update_comment(
            resource_type="aws.apigateway.domain_name", name=old_state["name"]
        )
        if not update_ret["result"]:
            result["result"] = False
            result["comment"] = update_ret["comment"]
        waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
            default_delay=15,
            default_max_attempts=40,
            timeout_config=timeout.get("update") if timeout else None,
        )
        update_waiter_acceptors = [
            {
                "matcher": "path",
                "expected": "AVAILABLE",
                "state": "success",
                "argument": "domainNameStatus",
            },
            {
                "matcher": "path",
                "expected": "UPDATING",
                "state": "retry",
                "argument": "domainNameStatus",
            },
            {
                "matcher": "path",
                "expected": "PENDING",
                "state": "retry",
                "argument": "domainNameStatus",
            },
            {
                "matcher": "path",
                "expected": "PENDING_CERTIFICATE_REIMPORT",
                "state": "retry",
                "argument": "domainNameStatus",
            },
            {
                "matcher": "path",
                "expected": "PENDING_OWNERSHIP_VERIFICATION",
                "state": "retry",
                "argument": "domainNameStatus",
            },
            {
                "matcher": "path",
                "expected": "NotFoundException",
                "state": "retry",
                "argument": "Error.Code",
            },
            {
                "matcher": "path",
                "expected": "BadRequestException",
                "state": "retry",
                "argument": "Error.Code",
            },
            {
                "matcher": "path",
                "expected": "ConflictException",
                "state": "retry",
                "argument": "Error.Code",
            },
            {
                "matcher": "path",
                "expected": "LimitExceededException",
                "state": "retry",
                "argument": "Error.Code",
            },
            {
                "matcher": "error",
                "expected": "UnauthorizedException",
                "state": "failure",
                "argument": "Error.Code",
            },
            {
                "matcher": "error",
                "expected": "TooManyRequestsException",
                "state": "retry",
                "argument": "Error.Code",
            },
        ]
        update_domain_waiter = hub.tool.boto3.custom_waiter.waiter_wrapper(
            name="ApiGatewayDomainNameUpdate",
            operation="GetDomainName",
            argument=["Error.Code", "domainNameStatus"],
            acceptors=update_waiter_acceptors,
            client=await hub.tool.boto3.client.get_client(ctx, "apigateway"),
        )
        try:
            await hub.tool.boto3.client.wait(
                ctx,
                "apigateway",
                "ApiGatewayDomainNameUpdate",
                update_domain_waiter,
                domainName=resource_id,
                WaiterConfig=waiter_config,
            )
        except Exception as e:
            result["comment"] = result["comment"] + (str(e),)
            result["result"] = False

    return result



from typing import Dict
from typing import List


async def apply(
    hub,
    ctx,
    resource,
    *,
    old_value: List[Dict[str, str]],
    new_value: List[Dict[str, str]],
    comments: List[str],
) -> bool:
    """
    This attribute can only be modified on creation of an instance
    """
    return True



"""Exec module for managing Amazon API Gateway Methods."""


async def get(
    hub,
    ctx,
    resource_id: str,
    name: str = None,
):
    # TODO: will revisit this part later to improve building the get() call when multiple params are used to fetch a resource from AWS
    """Get an API Gateway Method from AWS.

    Args:
        resource_id(str): Idem Resource ID that is generated once the resource is created.

        name(str, Optional): An Idem name of the API Gateway Method.
    """

    result = dict(comment=[], ret=None, result=True)
    get_method = await hub.exec.boto3.client.apigateway.get_method(
        ctx,
        restApiId=resource_id.split("-")[0],
        resourceId=resource_id.split("-")[1],
        httpMethod=resource_id.split("-")[2],
    )

    if not get_method["result"]:
        if "NotFoundException" in str(get_method["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.apigateway.method", name=name
                )
            )
            result["comment"] += list(get_method["comment"])
            return result
        result["comment"] += list(get_method["comment"])
        result["result"] = False
        return result
    if not get_method["ret"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.apigateway.method", name=name
            )
        )
        return result

    if get_method["ret"]:
        result["ret"] = hub.tool.aws.apigateway.method.convert_raw_method_to_present(
            raw_resource=get_method["ret"],
            idem_resource_name=name,
            resource_id=resource_id,
        )

    return result



import copy
import json
from collections import OrderedDict
from typing import Any
from typing import Dict


async def update_elasticsearch_domain(
    hub,
    ctx,
    resource_id: str,
    current_state: Dict[str, Any],
    desired_state: Dict[str, Any],
    timeout: dict = None,
):
    """Modifies the cluster configuration of the specified Elasticsearch domain. Also allows to upgrade
    domain version.

    Args:
        resource_id(str): The name of the Elasticsearch domain
        current_state(Dict): Previous state of the resource
        desired_state(Dict): New parameters from SLS file.
        timeout(dict, Optional): Timeout configuration for updating AWS Elasticsearch domain.

    Returns:
        Dict[str, Any]
    """
    update_params = dict(
        {
            "DomainName": "domain_name",
            "ElasticsearchClusterConfig": "elastic_search_cluster_config",
            "EBSOptions": "ebs_options",
            "AccessPolicies": "access_policies",
            "SnapshotOptions": "snapshot_options",
            "VPCOptions": "vpc_options",
            "CognitoOptions": "cognito_options",
            "EncryptionAtRestOptions": "encryption_at_rest_options",
            "NodeToNodeEncryptionOptions": "node_to_node_encryption_options",
            "AdvancedOptions": "advanced_options",
            "LogPublishingOptions": "log_publishing_options",
            "DomainEndpointOptions": "domain_endpoint_options",
            "AdvancedSecurityOptions": "advanced_security_options",
            "AutoTuneOptions": "auto_tune_options",
            "DryRun": "dry_run",
        }
    )
    upgrade_params = dict(
        {
            "TargetVersion": "elastic_search_version",
            "PerformCheckOnly": "perform_check_only",
        }
    )

    result = dict(comment=[], result=True, ret=None)
    plan_state = copy.deepcopy(current_state)
    params_to_update = {}
    params_to_upgrade = {}

    # create a dict 'params_to_modify' of raw parameter key mapped to desired value,
    # where the desired value is not none and current value does not match desired value
    for param_raw, param_present in update_params.items():
        if desired_state.get(param_present) is not None and current_state.get(
            param_present
        ) != desired_state.get(param_present):
            params_to_update[param_raw] = desired_state[param_present]

    for param_raw, param_present in upgrade_params.items():
        if desired_state.get(param_present) is not None and current_state.get(
            param_present
        ) != desired_state.get(param_present):
            params_to_upgrade[param_raw] = desired_state[param_present]

    if params_to_update or params_to_upgrade:
        for key, value in params_to_update.items():
            plan_state[update_params.get(key)] = value
        for key, value in params_to_upgrade.items():
            plan_state[upgrade_params.get(key)] = value
        if ctx.get("test", False):
            result["comment"] = hub.tool.aws.comment_utils.would_update_comment(
                resource_type="aws.es.elasticsearch_domain",
                name=current_state["name"],
            )
        else:
            result["comment"] = hub.tool.aws.comment_utils.update_comment(
                resource_type="aws.es.elasticsearch_domain",
                name=current_state["name"],
            )
            waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
                default_delay=30,
                default_max_attempts=60,
                timeout_config=timeout.get("update") if timeout else None,
            )
            if params_to_update:
                update_ret = (
                    await hub.exec.boto3.client.es.update_elasticsearch_domain_config(
                        ctx, DomainName=resource_id, **params_to_update
                    )
                )
                if not update_ret["result"]:
                    result["result"] = False
                    result["comment"] = update_ret["comment"]
                    return result

                update_waiter_acceptors = [
                    {
                        "matcher": "path",
                        "expected": False,
                        "state": "success",
                        "argument": "DomainStatus.Processing",
                    },
                    {
                        "matcher": "path",
                        "expected": True,
                        "state": "retry",
                        "argument": "DomainStatus.Processing",
                    },
                ]
                update_domain_waiter = hub.tool.boto3.custom_waiter.waiter_wrapper(
                    name="ElasticsearchDomainUpdate",
                    operation="DescribeElasticsearchDomain",
                    argument=["DomainStatus.Processing"],
                    acceptors=update_waiter_acceptors,
                    client=await hub.tool.boto3.client.get_client(ctx, "es"),
                )
                try:
                    await hub.tool.boto3.client.wait(
                        ctx,
                        "es",
                        "ElasticsearchDomainUpdate",
                        update_domain_waiter,
                        DomainName=resource_id,
                        WaiterConfig=waiter_config,
                    )
                except Exception as e:
                    result["comment"] = result["comment"] + (str(e),)
                    result["result"] = False
                    return result

            if params_to_upgrade:
                upgrade_ret = (
                    await hub.exec.boto3.client.es.upgrade_elasticsearch_domain(
                        ctx, DomainName=resource_id, **params_to_upgrade
                    )
                )
                if not upgrade_ret["result"]:
                    result["result"] = False
                    result["comment"] = upgrade_ret["comment"]
                    return result
                upgrade_waiter_acceptors = [
                    {
                        "matcher": "path",
                        "expected": False,
                        "state": "success",
                        "argument": "DomainStatus.UpgradeProcessing",
                    },
                    {
                        "matcher": "path",
                        "expected": True,
                        "state": "retry",
                        "argument": "DomainStatus.UpgradeProcessing",
                    },
                ]
                upgrade_domain_waiter = hub.tool.boto3.custom_waiter.waiter_wrapper(
                    name="ElasticsearchDomainUpgrade",
                    operation="DescribeElasticsearchDomain",
                    argument=["DomainStatus.UpgradeProcessing"],
                    acceptors=upgrade_waiter_acceptors,
                    client=await hub.tool.boto3.client.get_client(ctx, "es"),
                )
                try:
                    await hub.tool.boto3.client.wait(
                        ctx,
                        "es",
                        "ElasticsearchDomainUpgrade",
                        upgrade_domain_waiter,
                        DomainName=resource_id,
                        WaiterConfig=waiter_config,
                    )
                except Exception as e:
                    result["comment"] = result["comment"] + (str(e),)
                    result["result"] = False

        result["ret"] = plan_state
    else:
        result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
            resource_type="aws.es.elasticsearch_domain",
            name=current_state["name"],
        )
    return result


def convert_raw_elasticsearch_domain_to_present(
    hub,
    raw_resource: Dict[str, Any],
    idem_resource_name: str = None,
    tags: Dict[str, str] = None,
) -> Dict[str, Any]:
    """
    Convert AWS returned data structure to correct idem elasticsearch_domain present state

     Args:
        raw_resource(Dict): The aws response to convert
        idem_resource_name(str, Optional): An Idem name of the resource.
        tags(Dict, Optional): tags for Elasticsearch domain.

    Returns:
        Dict[str, Any]
    """
    domain_name = raw_resource.get("DomainName")
    resource_parameters = OrderedDict(
        {
            "DomainName": "domain_name",
            "ElasticsearchVersion": "elastic_search_version",
            "ElasticsearchClusterConfig": "elastic_search_cluster_config",
            "EBSOptions": "ebs_options",
            "AccessPolicies": "access_policies",
            "SnapshotOptions": "snapshot_options",
            "VPCOptions": "vpc_options",
            "CognitoOptions": "cognito_options",
            "EncryptionAtRestOptions": "encryption_at_rest_options",
            "NodeToNodeEncryptionOptions": "node_to_node_encryption_options",
            "AdvancedOptions": "advanced_options",
            "LogPublishingOptions": "log_publishing_options",
            "DomainEndpointOptions": "domain_endpoint_options",
            "AdvancedSecurityOptions": "advanced_security_options",
            "AutoTuneOptions": "auto_tune_options",
        }
    )
    resource_translated = {
        "name": idem_resource_name if idem_resource_name else domain_name,
        "resource_id": domain_name,
    }
    for parameter_raw, parameter_present in resource_parameters.items():
        if raw_resource.get(parameter_raw) is not None:
            resource_translated[parameter_present] = json.loads(
                json.dumps(raw_resource.get(parameter_raw))
            )
    if tags:
        resource_translated["tags"] = tags.copy()

    return resource_translated



from typing import Any
from typing import Dict


async def is_hosted_zone_associated_to_vpc(
    hub, ctx, zone_id: str, vpc_id: str, vpc_region: str
) -> Dict[str, Any]:
    """
    Check if a vpc is attached to a hosted zone

    Args:
        zone_id(str):
            The ID of the hosted zone to be associated with.

        vpc_id(str):
            The ID of the vpc you want to associate.

        vpc_region(str):
            The AWS region where the vpc belongs to.

    Returns:
        Dict[str, Any]

    """
    result = dict(comment=(), result=True, ret=None)
    ret_list = await hub.exec.boto3.client.route53.list_hosted_zones_by_vpc(
        ctx, VPCId=vpc_id, VPCRegion=vpc_region
    )
    if ret_list["result"]:
        if ret_list["ret"].get("HostedZoneSummaries"):
            associated_zone_summaries = ret_list["ret"].get("HostedZoneSummaries")
            hosted_zone_list = [
                hosted_zone_summary.get("HostedZoneId")
                for hosted_zone_summary in associated_zone_summaries
            ]
            if zone_id in hosted_zone_list:
                result["ret"] = {"HostedZone": zone_id}
    else:
        result["result"] = False
        result["comment"] = ret_list["comment"]
    return result



"""State module for managing Amazon ECR Repository Policies."""
import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    repository_name: str,
    policy_text: Dict or str,
    resource_id: str = None,
    registry_id: str = None,
    force: bool = False,
) -> Dict[str, Any]:
    """Applies a repository policy to the specified AWS ECR repository to control access permissions.

    Args:
        name(str):
            An Idem name of the resource.
        repository_name(str):
            The name of the ECR repository in Amazon Web Services to receive the policy.
        policy_text(dict or str):
            The JSON repository policy text to apply to the repository.
        resource_id(str, Optional):
            The registry id and repository name with a separator '-'. Format: ``[registry_id]-[repository_name]``.
        registry_id(str, Optional):
            The Amazon Web Services account ID associated with the registry that contains the repository. If you
            do not specify a registry, the default registry is assumed.
        force(bool, Optional):
            If the policy you are attempting to set on a repository policy would prevent you from setting another policy
            in the future, you must force the SetRepositoryPolicy operation. This is intended to prevent accidental repository lock outs.

    Request Syntax:
      .. code-block:: sls

        [idem_test_aws_ecr_repository_policy]:
          aws.ecr.repository_policy.present:
            - name: 'string'
            - repository_name: 'string'
            - policy_text: 'string'
            - registry_id: 'string'
            - force: True|False

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            idem_test_aws_ecr_repository_policy:
              aws.ecr.repository.present:
                - name: idem_test_ecr_policy
                - repository_name: idem_test_ecr_repository
                - policy_text:
                    Version: '2012-10-17'
                    Statement:
                      - Sid: 'DenyPull'
                        Effect: 'Deny'
                        Principal: '*'
                        Action: ['ecr:BatchGetImage', 'ecr:GetDownloadUrlForLayer']
                - registry_id: idem_test_ecr_registry
                - force: True
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    resource_updated = False
    policy_text = hub.tool.aws.state_comparison_utils.standardise_json(policy_text)

    if resource_id:
        if registry_id is None:
            registry_id = resource_id.split("-")[0]

    before_ret = await hub.exec.aws.ecr.repository_policy.get(
        ctx,
        name=repository_name,
        registry_id=registry_id,
        resource_id=repository_name,
    )

    # Case: If there is something found
    if before_ret["result"] and before_ret["ret"]:
        result["comment"] = (f"aws.ecr.repository_policy '{name}' already exists",)
        result["old_state"] = copy.deepcopy(before_ret.get("ret"))
        result["new_state"] = copy.deepcopy(before_ret.get("ret"))

        if policy_text and not hub.tool.aws.state_comparison_utils.is_json_identical(
            result["old_state"]["policy_text"], policy_text
        ):
            if ctx.get("test", False):
                result["comment"] = result["comment"] + (
                    f"Would update policy_text for aws.ecr.repository_policy '{name}'",
                )
                result["new_state"]["policy_text"] = policy_text
                return result

            update_ret = await hub.exec.boto3.client.ecr.set_repository_policy(
                ctx,
                registryId=registry_id,
                repositoryName=repository_name,
                policyText=policy_text,
                force=force,
            )
            if not update_ret["result"]:
                result["result"] = False
                result["comment"] = result["comment"] + update_ret["comment"]
                return result

            result["comment"] = result["comment"] + (
                f"Updated aws.ecr.repository_policy '{name}'",
            )
            resource_updated = True
    else:
        if ctx.get("test", False):
            result[
                "new_state"
            ] = raw_resource = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "resource_id": f"{registry_id}-{repository_name}",
                    "registry_id": registry_id,
                    "repository_name": repository_name,
                    "policy_text": policy_text,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.ecr.repository_policy", name=name
            )
            return result

        set_policy_ret = await hub.exec.boto3.client.ecr.set_repository_policy(
            ctx,
            registryId=registry_id,
            repositoryName=repository_name,
            policyText=policy_text,
            force=force,
        )
        if not set_policy_ret["result"]:
            result["result"] = False
            result["comment"] = set_policy_ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.ecr.repository_policy", name=name
        )

    # Case: If something got updated or old_state is not present
    if (not result["old_state"]) or resource_updated:
        after_ret = await hub.exec.aws.ecr.repository_policy.get(
            ctx, name=name, registry_id=registry_id, resource_id=repository_name
        )

        if not after_ret["result"]:
            result["result"] = False
            result["comment"] = result["comment"] + after_ret["comment"]
            return result

        result["new_state"] = copy.deepcopy(after_ret.get("ret"))

    return result


async def absent(
    hub,
    ctx,
    name: str,
    repository_name: str,
    resource_id: str = None,
    registry_id: str = None,
) -> Dict[str, Any]:
    """Deletes the policy associated with the specified AWS ECR repository.

    Args:
        name(str):
            An Idem name of the resource.
        repository_name(str):
            The name of the ECR repository in Amazon Web Services that contains the policy to delete.
        resource_id(str, Optional):
            The registry id and repository name with a separator '-'. Format: ``[registry_id]-[repository_name]``.

            .. warning::
              Idem automatically considers this resource being absent if this field is not specified.

        registry_id(str, Optional):
            The Amazon Web Services account ID associated with the registry that contains the repository policy to
            delete. If you do not specify a registry, the default registry is assumed.

    Request syntax:
      .. code-block:: sls

        [idem_test_aws_ecr_repository_policy]:
          aws.ecr.repository_policy.absent:
            - name: 'string'
            - repository_name: 'string'
            - resource_id: 'string'
            - registry_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            idem_test_aws_ecr_repository_policy:
              aws.ecr.repository_policy.absent:
                - name: idem_test_ecr_policy
                - repository_name: idem_test_ecr_repository
                - resource_id: idem_test_ecr_registry-idem_test_ecr_repository
                - registry_id: idem_test_ecr_registry
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ecr.repository_policy", name=name
        )
        return result
    else:
        if registry_id is None:
            registry_id = resource_id.split("-")[0]

    before_ret = await hub.exec.aws.ecr.repository_policy.get(
        ctx,
        name=name,
        registry_id=registry_id,
        resource_id=repository_name,
    )

    # Case: Error
    if not before_ret["result"]:
        result["result"] = False
        result["comment"] = before_ret["comment"]
        return result

    # Any case forward, the old_state should carry current state in old_state
    result["old_state"] = copy.deepcopy(before_ret["ret"])

    if not before_ret["ret"]:
        # Case: already absent
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ecr.repository_policy", name=name
        )
        return result
    elif ctx.get("test", False):
        # Case: --test context
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.ecr.repository_policy", name=name
        )
        return result

    delete_ret = await hub.exec.boto3.client.ecr.delete_repository_policy(
        ctx, registryId=registry_id, repositoryName=repository_name
    )

    # Case: Failed to delete
    if not delete_ret["result"]:
        result["result"] = False
        result["comment"] = delete_ret["comment"]
        return result

    # Case: All good, add deleted comment
    result["comment"] = hub.tool.aws.comment_utils.delete_comment(
        resource_type="aws.ecr.repository_policy", name=name
    )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describes policies for AWS ECR repositories in a registry in a way that can be recreated/managed with the corresponding "present" function.

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:
        .. code-block:: bash

            $ idem describe aws.ecr.repository_policy
    """
    result = {}

    # To describe all the repository policies, we first need to list all the repositories,
    # and then we get the policy associated to each repository
    describe_ret = await hub.exec.aws.ecr.repository.list(ctx)
    if not describe_ret["result"]:
        hub.log.warning(f"Could not describe repositories: {describe_ret['comment']}")
        return result

    for repository in describe_ret["ret"]:
        registry_id = repository.get("registry_id")
        repository_name = repository.get("name")

        get_policy_ret = await hub.exec.aws.ecr.repository_policy.get(
            ctx,
            name=repository_name,
            registry_id=registry_id,
            resource_id=repository_name,
        )

        # Case: failure to get a policy
        if not get_policy_ret["result"]:
            hub.log.warning(
                f"Could not get policy for repository '{registry_id}/{repository_name}': "
                f"{get_policy_ret['comment']}. Describe will skip this repository and continue."
            )
            continue

        # Case: There is no policy for the repository
        if not get_policy_ret.get("ret"):
            hub.log.debug(
                f"No policy for repository '{registry_id}/{repository_name}' exists: "
                f"{get_policy_ret['comment']}. Describe will skip this repository and continue."
            )
            continue

        resource = get_policy_ret.get("ret")
        result[resource.get("name")] = {
            "aws.ecr.repository_policy.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }

    return result
